# Setup script.
#
# Author: Malte J. Ziebarth (mjz.science@fmvkb.de)
#
# Copyright (C) 2024 Malte J. Ziebarth
#
# Licensed under the EUPL, Version 1.2 or – as soon they will be approved by
# the European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the Licence for the specific language governing permissions and
# limitations under the Licence.

from pathlib import Path
from shutil import copyfile
from setuptools import setup, Extension
from mebuex import MesonExtension, build_ext

unit     = MesonExtension('cyantities.unit')
quantity = MesonExtension('cyantities.quantity')

#
# Post compile
# ------------
# Get the compiled static library for linking.
#
class InstallStaticLibrary(build_ext):
    """
    """
    def run(self):
        # First compile:
        super().run()

        # Copy the static library:
        builddir = Path('.') / "builddir"
        targetdir = quantity.mebuex_destpath
        copyfile(
            builddir / "libcyantities.a",
            targetdir / "_cpp" / "cyantities" / "libcyantities.a"
        )


setup(ext_modules=[unit, quantity],
      cmdclass={'build_ext' : InstallStaticLibrary}
)